import re

################################################################################
# Functions for fasta formatted file parsing and validation
################################################################################
def isFasta(fileInLines):
    '''
    Input: List of lines of a stream. A line is defined as a sequence of characters that does not contain 
    newline character anywhere but at the very end.
    
    Function: Checks if the given stream is a valid fasta file. Valid fasta file contains 
    a number of protein sequences that could extend to more than 1 line separated by description lines
    that start with '>' character. Description line should come before corresponding protein sequence.
    
    Output: Either a non-empty string containing the error message or an empty string indicating successful 
    validation is returned.
    '''
    DESCRIPTION = 1
    SEQUENCE = 2
    state = 0 # state variable is initialized to zero and will be taking values DESCRIPTION and SEQUENCE during the processing of the given stream
    
    # Validation of the given stream to make sure there are no missing description lines or 
    # corresponding sequences, and these are in proper, alternating order 
    for line in fileInLines:
        if(line == ""): #ignore blank lines
            pass
        elif(line[0] == ">"): #Then, this line must be a description line
            if state == DESCRIPTION:
                return "Not a fasta file: Missing sequence"    
            state = DESCRIPTION
        else: #Then, this line must be part of a sequence
            if state == SEQUENCE or state == DESCRIPTION:
                state = SEQUENCE
            else:
                return "Not a fasta file: Missing description line"
    
    if state != SEQUENCE:
        return "Not a fasta file: Missing sequence"
        
    return ""
 
def extractIdFromDescriptionLine(line):
    '''
    Input: A fasta description line that starts with the character '>'
    
    Function: Assuming either of the following two scenarios, extracts id from the given line.
              - id follows '>' character in a line and either it is followed by nothing or it
              is followed by '|' character marking the end of the id and the beginning of the
              protein description (which will be ignored).
              - line starts with the character sequence '>gi|' and id follows this, '|' character 
              marking the end of the id
              
    Output: Returns the extracted id as a string          
    '''
    giPattern = re.compile('>gi|')
    #extract protein id from the description line
    matchObj = re.search('>gi\\|', line)
    if(matchObj is not None):
        line = line[matchObj.end():]
        record = line.split("|"); 
        return record[0].strip()
    else:
        return line[1:].split('|')[0].strip()
          
def fetchIDandSequence(fastaInLines):
    ''' 
    Input: List of lines of a stream containing a valid fasta formatted file. Assumes a valid fasta formatted 
    file is given as input (check isFasta function above to learn what valid fasta format is supposed to be)
    
    Function: Parses lines of a fasta formated file to get the list of IDs and sequences of proteins
    
    Output: Parallel lists of ids and protein sequence contained in the given stream
    '''
    DESCRIPTION = 1
    SEQUENCE = 2
    state = 0 # state variable is initialized to zero and will be taking values DESCRIPTION and SEQUENCE during the processing of the given stream
    idList = []
    sequences=[]
    seq = ""  
    for line in fastaInLines:
        if(line == ""): #ignore blank lines
            pass
        elif(line[0] == ">"): #Then, this line must be a description line        
            idList.append(extractIdFromDescriptionLine(line))   
            if state == SEQUENCE:
                sequences.append(seq)
            state = DESCRIPTION
        else: #Then, this line must be part of a sequence
            if (state == DESCRIPTION): # beginning of a new sequence
                state = SEQUENCE
                seq = ""  
            seq += line.strip()
    sequences.append(seq) #Add last sequence to the list
    return sequences, idList

